from typing import Annotated, Union

from pydantic import Field

from .allwise_p3as_mep import allwise_p3as_mep
from .allwise_p3as_psd import allwise_p3as_psd
from .neowiser_p1bs_psd import neowiser_p1bs_psd

TableType = Annotated[
    Union[allwise_p3as_mep, neowiser_p1bs_psd, allwise_p3as_psd], Field(discriminator="name")
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/timewise/t2/T2StackVisits.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                24.09.2025
# Last Modified Date:  24.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>
from typing import Literal
from scipy import stats

from ampel.abstract.AbsLightCurveT2Unit import AbsLightCurveT2Unit
from ampel.struct.UnitResult import UnitResult
from ampel.types import UBson
from ampel.view.LightCurve import LightCurve

from ampel.timewise.util.pdutil import datapoints_to_dataframe

from timewise.process import keys
from timewise.process.stacking import stack_visits


SIGMA = float(stats.chi2.cdf(1, 1))


class T2StackVisits(AbsLightCurveT2Unit):
    clean_outliers: bool = True

    # spread quantile of each visit to estimate, defaults to
    # ~68% which is the equivalent of 1 sigma for a normal distribution
    outlier_quantile: float = SIGMA

    # threshold above which to exclude outliers
    outlier_threshold: float = 5

    # methods to calculate mean and std
    mean_name: Literal["mean", "median"] = "median"
    std_name: Literal["std", "sdom", "sdom-1"] = "sdom-1"
    correction_name: Literal["debias", "tdist", "none"] = "tdist"

    # see timewise.process.stacking
    calculate_pvalues: bool = False
    use_single_exposure_errors: bool = True
    median_zeropoint_per_visit: bool = True

    def process(self, light_curve: LightCurve) -> UBson | UnitResult:
        columns = [
            "ra",
            "dec",
            "mjd",
        ]
        for i in range(1, 3):
            for key in [keys.MAG_EXT, keys.FLUX_EXT]:
                columns.extend([f"w{i}{key}", f"w{i}{keys.ERROR_EXT}{key}"])

        photopoints = light_curve.get_photopoints()
        if photopoints is None:
            return {}
        data, _ = datapoints_to_dataframe(photopoints, columns=columns)
        if len(data) == 0:
            return {}
        return stack_visits(
            data,
            outlier_threshold=self.outlier_threshold,
            outlier_quantile=self.outlier_quantile,
            clean_outliers=self.clean_outliers,
            mean_name=self.mean_name,
            std_name=self.std_name,
            correction_name=self.correction_name,
            calculate_pvalues=self.calculate_pvalues,
            use_single_exposure_errors=self.use_single_exposure_errors,
            median_zeropoint_per_visit=self.median_zeropoint_per_visit,
        ).to_dict(orient="records")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                timewise/ampel/timewise/t1/TimewiseFilter.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                16.09.2025
# Last Modified Date:  16.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>
import numpy as np

from ampel.abstract.AbsAlertFilter import AbsAlertFilter
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol

from timewise.util.visits import get_visit_map
from timewise.process import keys


class TimewiseFilter(AbsAlertFilter):
    det_per_visit: int = 8
    n_visits: int = 10

    def process(self, alert: AmpelAlertProtocol) -> None | bool | int:
        columns = [
            "ra",
            "dec",
            "mjd",
        ]
        for i in range(1, 3):
            for key in [keys.MAG_EXT, keys.FLUX_EXT]:
                columns.extend([f"w{i}{key}", f"w{i}{keys.ERROR_EXT}{key}"])

        mjd = np.array([dp["mjd"] for dp in alert.datapoints])

        # enough single detections per visit
        visit_map = get_visit_map(mjd)
        visits, counts = np.unique(visit_map, return_counts=True)
        visit_passed = counts >= self.det_per_visit
        if not all(visit_passed):
            self.logger.debug(None, extra={"min_det_per_visit": min(counts).item()})
            return None

        # enough visits
        if not len(visits) >= self.n_visits:
            self.logger.debug(None, extra={"n_visits": len(visits)})
            return None

        return True



from pathlib import Path
import os


def expand(path: Path | str) -> Path:
    """
    Fully expand and resolve the Path with the given environment variables.
    """
    path = Path(path)
    return Path(os.path.expandvars(path)).expanduser().resolve()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/timewise/t1/T1HDBSCAN.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                24.09.2025
# Last Modified Date:  24.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>
from typing import Iterable, Sequence

import numpy as np
from numpy import typing as npt
from sklearn.cluster import HDBSCAN
from pymongo import MongoClient

from ampel.content.DataPoint import DataPoint
from ampel.struct.T1CombineResult import T1CombineResult
from ampel.types import DataPointId
from ampel.abstract.AbsT1CombineUnit import AbsT1CombineUnit
from ampel.model.UnitModel import UnitModel
from ampel.base.AuxUnitRegister import AuxUnitRegister

from ampel.timewise.util.pdutil import datapoints_to_dataframe
from ampel.timewise.util.AuxDiagnosticPlotter import AuxDiagnosticPlotter
from timewise.process import keys

from importlib.util import find_spec

if find_spec("astropy.coordinates.angle_utilities"):
    # astropy < v6.0.0
    from astropy.coordinates.angle_utilities import angular_separation, position_angle
else:
    # astropy >= v6.0.0
    from astropy.coordinates.angles import angular_separation, position_angle


class T1HDBSCAN(AbsT1CombineUnit):
    input_mongo_db_name: str
    original_id_key: str
    mongo: str = "mongodb://localhost:27017/"
    whitelist_region_arcsec: float = 1
    cluster_distance_arcsec: float = 0.5

    plot: bool = False
    plotter: UnitModel = UnitModel.model_validate(
        {
            "unit": "AuxDiagnosticPlotter",
            "config": {
                "plot_properties": {
                    "file_name": {
                        "format_str": "%s_hdbscan_selection.svg",
                        "arg_keys": ["stock"],
                    }
                }
            },
        }
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._col = MongoClient(self.mongo)[self.input_mongo_db_name]["input"]
        self._plotter = AuxUnitRegister.new_unit(
            model=self.plotter, sub_type=AuxDiagnosticPlotter
        )

    def combine(
        self, datapoints: Iterable[DataPoint]
    ) -> Sequence[DataPointId] | T1CombineResult:
        columns = ["ra", "dec", "mjd"]
        if self.plot:
            for i in range(1, 3):
                columns += [
                    f"w{i}{self._plotter.lum_key}",
                    f"w{i}{keys.ERROR_EXT}{self._plotter.lum_key}",
                ]

        lightcurve, stock_ids = datapoints_to_dataframe(
            datapoints, columns, check_tables=["allwise_p3as_mep"]
        )

        # make sure that the is one stock id that fits all dps
        # this is a redundant check, the muxer should take care of it
        unique_stocks = np.unique(np.array(stock_ids).flatten())
        stock_in_all_dps = [
            all([s in sids for sids in stock_ids]) for s in unique_stocks
        ]
        # make sure only one stock is in all datapoints
        assert sum(stock_in_all_dps) == 1
        stock_id = unique_stocks[stock_in_all_dps][0].item()
        self.logger.debug(f"stock: {stock_id}")

        # query the database that holds the parent sample
        d = self._col.find_one({self.original_id_key: stock_id})
        source_ra = d["ra"]
        source_dec = d["dec"]

        lc_ra_rad = np.deg2rad(lightcurve.ra)
        lc_dec_rad = np.deg2rad(lightcurve.dec)
        source_ra_rad = np.deg2rad(source_ra)
        source_dec_rad = np.deg2rad(source_dec)

        # calculate separation and position angle
        _angular_separation = angular_separation(
            source_ra_rad, source_dec_rad, lc_ra_rad, lc_dec_rad
        )
        _position_angle = position_angle(
            source_ra_rad, source_dec_rad, lc_ra_rad, lc_dec_rad
        )

        # The AllWISE multiframe pipeline detects sources on the deep coadded atlas images and then measures the sources
        # for all available single-exposure images in all bands simultaneously, while the NEOWISE magnitudes are
        # obtained by PSF fit to individual exposures directly. Effect: all allwise data points that belong to the same
        # object have the same position. We take only the closest one and treat it as one datapoint in the clustering.
        allwise_time_mask = lightcurve.allwise_p3as_mep
        if any(allwise_time_mask):
            allwise_sep_min = np.min(_angular_separation[allwise_time_mask])
            closest_allwise_mask = (
                _angular_separation == allwise_sep_min
            ) & allwise_time_mask
            closest_allwise_mask_first_entry = (
                ~closest_allwise_mask.duplicated() & closest_allwise_mask
            )

            # the data we want to use is then the selected AllWISE datapoint and the NEOWISE-R data
            data_mask = closest_allwise_mask_first_entry | ~allwise_time_mask
        else:
            closest_allwise_mask_first_entry = closest_allwise_mask = None
            data_mask = np.ones_like(_angular_separation, dtype=bool)

        # no matter which cluster they belong to, we want to keep all datapoints within 1 arcsec
        one_arcsec_mask = _angular_separation < np.radians(
            self.whitelist_region_arcsec / 3600
        )
        selected_indices = set(lightcurve.index[data_mask & one_arcsec_mask])

        # if there are more than one datapoints, we use a clustering algorithm to potentially find a cluster with
        # its center within 1 arcsec
        if data_mask.sum() > 1:
            # instead of the polar coordinates separation and position angle we use cartesian coordinates because the
            # clustering algorithm works better with them
            cartesian_full = np.array(
                [
                    _angular_separation * np.cos(_position_angle),
                    _angular_separation * np.sin(_position_angle),
                ]
            ).T
            cartesian = cartesian_full[data_mask]

            # we are now ready to do the clustering
            cluster_res = HDBSCAN(
                store_centers="centroid",
                min_cluster_size=max(min(20, len(cartesian)), 2),
                allow_single_cluster=True,
                cluster_selection_epsilon=np.radians(
                    self.cluster_distance_arcsec / 3600
                ),
            ).fit(cartesian)
            centroids: npt.NDArray[np.float64] = cluster_res.__getattribute__(
                "centroids_"
            )
            labels: npt.NDArray[np.int64] = cluster_res.__getattribute__("labels_")

            # we select the closest cluster within 1 arcsec
            cluster_separations = np.sqrt(np.sum(centroids**2, axis=1))
            self.logger.debug(f"Found {len(cluster_separations)} clusters")

            # if there is no cluster or no cluster within 1 arcsec,
            # only the datapoints within 1 arcsec are selected as we did above
            if len(cluster_separations) == 0:
                self.logger.debug(
                    "No cluster found. Selecting all noise datapoints within 1 arcsec."
                )
            elif (_min := min(cluster_separations)) > np.radians(
                self.whitelist_region_arcsec / 3600
            ):
                self.logger.debug(
                    f"Closest cluster is at {np.degrees(_min) * 3600} arcsec"
                )

            # if there is a cluster within 1 arcsec, we select all datapoints belonging to that cluster
            # in addition to the datapoints within 1 arcsec
            else:
                closest_label = cluster_separations.argmin()
                selected_cluster_mask = labels == closest_label

                # now we have to trace back the selected datapoints to the original lightcurve
                selected_indices |= set(
                    lightcurve.index[data_mask][selected_cluster_mask]
                )
                self.logger.debug(f"Selected {len(selected_indices)} datapoints")

        else:
            # if there is only one datapoint we give him a label manually
            labels = np.array([-1])

        # if the closest allwise source is selected, we also select all other detections belonging to that
        # source in the allwise period
        if (
            closest_allwise_mask_first_entry is not None
            and lightcurve.index[closest_allwise_mask_first_entry][0]
            in selected_indices
        ):
            closest_allwise_mask_not_first = (
                closest_allwise_mask & ~closest_allwise_mask_first_entry
            )
            closest_allwise_indices_not_first = lightcurve.index[
                closest_allwise_mask_not_first
            ]
            self.logger.debug(
                f"Adding remaining {len(closest_allwise_indices_not_first)} from AllWISE period"
            )
            selected_indices |= set(closest_allwise_indices_not_first)

        selected_indices_list = list(selected_indices)
        res = T1CombineResult(dps=selected_indices_list)

        if self.plot:
            all_labels = np.array([-1] * len(lightcurve))
            all_labels[data_mask] = labels
            svg_rec = self._plotter.make_plot(
                lightcurve,
                None,
                all_labels,
                source_ra,
                source_dec,
                selected_indices_list,
                highlight_radius=self.whitelist_region_arcsec,
            )
            res.add_meta("plot", svg_rec)

        return res



#!/usr/bin/env python
# File:                timewise/ampel/timewise/ingest/tags.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                19.09.2025
# Last Modified Date:  19.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>


# tags is used by TiT0PhotoPointShaper and ZiT0UpperLimitShaper
# key: filter id
tags: dict[str, list[str]] = {
    "allwise_p3as_mep": ["WISE", "TIMEWISE", "allwise_p3as_mep"],
    "neowiser_p1bs_psd": ["WISE", "TIMEWISE", "neowiser_p1bs_psd"],
}



#!/usr/bin/env python
# File:                timewise/ampel/timewise/ingest/TiDataPointShaper.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  19.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>

from collections.abc import Iterable, Sequence
from typing import Any

from bson import encode

from ampel.abstract.AbsT0Unit import AbsT0Unit
from ampel.base.AmpelUnit import AmpelUnit
from ampel.content.DataPoint import DataPoint
from ampel.types import StockId, Tag
from ampel.util.hash import hash_payload

from ampel.timewise.ingest.tags import tags


class TiDataPointShaperBase(AmpelUnit):
    """
    This class 'shapes' datapoints in a format suitable
    to be saved into the ampel database
    """

    # JD2017 is used to define upper limits primary IDs
    JD2017: float = 2457754.5
    #: Byte width of datapoint ids
    digest_size: int = 8

    # Mandatory implementation
    def process(self, arg: Iterable[dict[str, Any]], stock: StockId) -> list[DataPoint]:
        """
        :param arg: sequence of unshaped pps
        IMPORTANT:
        1) This method *modifies* the input dicts (it removes 'candid' and programpi),
        even if the unshaped pps are ReadOnlyDict instances
        2) 'stock' is not set here on purpose since it will conflict with the $addToSet operation
        """

        ret_list: list[DataPoint] = []
        popitem = dict.pop

        for photo_dict in arg:
            # Photopoint
            assert photo_dict.get("candid"), "photometry points does not have 'candid'!"
            ret_list.append(
                {  # type: ignore[typeddict-item]
                    "id": photo_dict["candid"],
                    "stock": stock,
                    "tag": tags[photo_dict["table_name"]],
                    "body": photo_dict,
                }
            )

            popitem(photo_dict, "candid", None)

        return ret_list

    def _create_datapoint(
        self, stock: StockId, tag: Sequence[Tag], body: dict[str, Any]
    ) -> DataPoint:
        """
        Create a Datapoint from stock, body, and tags, using the hash of the body as id
        """
        # ensure that keys are ordered
        sorted_body = dict(sorted(body.items()))
        # The following is a comment from the original ampel.ztf.ingest.ZiDataPointShaperBase:
        # This is not a complete DataPoint as (channel,meta) is missing, set later.
        # Should these be optional? or added default?
        return {  # type: ignore
            "id": hash_payload(encode(sorted_body), size=-self.digest_size * 8),
            "stock": stock,
            "tag": [*tags[body["table_name"]], *tag],
            "body": sorted_body,
        }

    def ul_identity(self, uld: dict[str, Any]) -> int:
        """
        This should not happen
        """
        raise NotImplementedError


class TiDataPointShaper(TiDataPointShaperBase, AbsT0Unit):
    def process(self, arg: Any, stock: None | StockId = None) -> list[DataPoint]:
        assert stock is not None
        return super().process(arg, stock)



from typing import Literal, Dict, Any

import pandas as pd
from numpy import typing as npt

from ampel.plot.create import create_plot_record
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AmpelABC import AmpelABC
from ampel.model.PlotProperties import PlotProperties
from ampel.content.NewSVGRecord import NewSVGRecord

from timewise.plot.diagnostic import DiagnosticPlotter


class AuxDiagnosticPlotter(AmpelBaseModel, AmpelABC):
    plot_properties: PlotProperties
    cutout: Literal["sdss", "panstarrs"] = DiagnosticPlotter.model_fields[
        "cutout"
    ].default
    band_colors: Dict[str, str] = DiagnosticPlotter.model_fields["band_colors"].default
    lum_key: str = DiagnosticPlotter.model_fields["lum_key"].default

    def __init__(self, /, **data: Any):
        super().__init__(**data)
        self._plotter = DiagnosticPlotter.model_validate(
            {k: self.__getattribute__(k) for k in ["cutout", "band_colors", "lum_key"]}
        )

    def make_plot(
        self,
        raw_lightcurve: pd.DataFrame,
        stacked_lightcurve: pd.DataFrame | None,
        labels: npt.ArrayLike,
        source_ra: float,
        source_dec: float,
        selected_indices: list[Any],
        highlight_radius: float | None = None,
    ) -> NewSVGRecord:
        fig, axs = self._plotter.make_plot(
            raw_lightcurve=raw_lightcurve,
            stacked_lightcurve=stacked_lightcurve,
            labels=labels,
            source_ra=source_ra,
            source_dec=source_dec,
            selected_indices=selected_indices,
            highlight_radius=highlight_radius,
        )
        return create_plot_record(fig, self.plot_properties)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                timewise/ampel/timewise/alert/load/TimewiseFileLoader.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                16.09.2025
# Last Modified Date:  16.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>
from typing import Dict, get_args

import numpy as np
import pandas as pd
from astropy.table import vstack
from ampel.abstract.AbsAlertLoader import AbsAlertLoader
from ampel.base.AmpelABC import AmpelABC
from timewise.tables import TableType
from timewise.config import TimewiseConfig
from timewise.types import TaskID
from timewise.util.path import expand


class TimewiseFileLoader(AbsAlertLoader[Dict], AmpelABC):
    """
    Load alerts from one of more files.
    """

    # path to timewise download config file
    timewise_config_file: str

    # column name of id
    stock_id_column_name: str

    chunks: list[int] | None = None

    # optionally skip files that are missing
    skip_missing_files: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        expanded_config_file = expand(self.timewise_config_file)

        self.logger.info(f"loading timewise config file {expanded_config_file}")
        timewise_config = TimewiseConfig.from_yaml(expanded_config_file)
        dl = timewise_config.download.build_downloader()
        self._timewise_backend = dl.backend

        # selecting tasks to run
        _tasks = [tasks for tasks in dl.iter_tasks_per_chunk()]
        if self.chunks is not None:
            self._tasks = [_tasks[i] for i in self.chunks]
        else:
            self._tasks = _tasks
        if self.logger:
            self.logger.info(
                f"Registering {len(self._tasks)} chunk(s) to load: {self._tasks}"
            )

        self._table_types = get_args(TableType.__origin__)  # type: ignore
        self._gen = self.iter_stocks()

    @staticmethod
    def encode_result(res: pd.DataFrame) -> pd.DataFrame:
        if isinstance(res, pd.Series):
            return pd.DataFrame([res])
        return res

    def find_table_from_task(self, task: TaskID) -> TableType:
        tables = [
            t for t in self._table_types if t.model_fields["name"].default in str(task)
        ]
        assert len(tables) > 0, f"No matching table found for {task}!"
        assert len(tables) < 2, f"More than one matching table found for {task}!"
        self.logger.debug(
            f"{task} is from table {tables[0].model_fields['name'].default}"
        )
        return tables[0]

    def iter_stocks(self):
        # emit all datapoints per stock id
        # This way ampel runs not per datapoint but per object
        backend = self._timewise_backend
        for tasks in self._tasks:
            data = []
            for task in tasks:
                self.logger.debug(f"reading {task}")
                try:
                    idata = backend.load_data(task)
                except FileNotFoundError as e:
                    if self.skip_missing_files:
                        self.logger.warn(f"file for task {task} not found, skipping...")
                        continue
                    else:
                        raise e

                # add table name
                idata["table_name"] = (
                    self.find_table_from_task(task).model_fields["name"].default
                )

                data.append(idata)

            data = vstack(data).to_pandas()

            # rename stock id column
            data.rename(columns={self.stock_id_column_name: "stock_id"}, inplace=True)

            # Find the indices for each stock id. This is much faster than making a mask
            # each loop and accessing the table then. Shown below is a comparison.
            # The top example is the access provided by pandas which would be
            # again a factor 3 faster.
            #
            # In [45]: %timeit test_df()
            # 5.62 μs ± 47.2 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
            #
            # In [46]: %timeit test_index()
            # 14.6 μs ± 45 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
            #
            # In [47]: %timeit test_mask()
            # 2.61 ms ± 18 μs per loop (mean ± std. dev. of 7 runs, 100 loops each)
            data.set_index(data.stock_id, inplace=True)

            # iterate over all stock ids
            for stock_id in np.unique(data["stock_id"]):
                selection = data.loc[stock_id]
                yield self.encode_result(selection)

    def __iter__(self):
        return self

    def __next__(self) -> pd.DataFrame:  # type: ignore
        return next(self._gen)



import logging
from functools import cached_property
from pathlib import Path
from typing import Iterator, Sequence, cast

import numpy as np
import pandas as pd
from numpy import typing as npt

logger = logging.getLogger(__name__)


class Chunk:
    def __init__(self, chunk_id: int, input_csv, row_indices: npt.NDArray[np.int_]):
        self.chunk_id = chunk_id
        self.row_numbers = row_indices
        self.input_csv = input_csv

    @cached_property
    def indices(self) -> pd.Index:
        start = min(self.row_numbers)
        stop = max(self.row_numbers) + 1
        return pd.read_csv(self.input_csv, skiprows=start, nrows=stop - start).index

    @property
    def data(self) -> pd.DataFrame:
        start = (
            min(cast(Sequence[int], self.row_numbers)) + 1
        )  # plus one to always skip header line
        nrows = (
            max(cast(Sequence[int], self.row_numbers)) - start + 2
        )  # plus one: skip header, plus one:

        columns = list(pd.read_csv(self.input_csv, nrows=0).columns)
        return pd.read_csv(
            filepath_or_buffer=self.input_csv,
            skiprows=start,
            nrows=nrows,
            names=columns,
        )


class Chunker:
    def __init__(self, input_csv: Path, chunk_size: int):
        self.input_csv = input_csv
        self.chunk_size = chunk_size
        self._n_rows = self._count_rows()
        logger.debug(f"found {self._n_rows} rows in {self.input_csv}")

    def _count_rows(self) -> int:
        chunk = 1024 * 1024  # Process 1 MB at a time.
        f = np.memmap(self.input_csv)
        num_newlines = sum(
            np.sum(f[i : i + chunk] == ord("\n")) for i in range(0, len(f), chunk)
        )
        del f
        return num_newlines - 1  # one header row

    def __len__(self) -> int:
        return int(np.ceil(self._n_rows / self.chunk_size))

    def __iter__(self) -> Iterator[Chunk]:
        for chunk_id in range(len(self)):
            yield self.get_chunk(chunk_id)

    def get_chunk(self, chunk_id: int) -> Chunk:
        if chunk_id >= len(self):
            raise IndexError(f"Invalid chunk_id {chunk_id}")
        start = chunk_id * self.chunk_size
        stop = min(start + self.chunk_size, self._n_rows)
        logger.debug(f"chunk {chunk_id}: from {start} to {stop}")
        return Chunk(chunk_id, self.input_csv, np.arange(start, stop))



#!/usr/bin/env python
# File:                ampel/timewise/ingest/TiMongoMuxer.py
# License:             BSD-3-Clause
# Author:              Jannis Necker
# Date:                19.09.2025
# Last Modified Date:  27.09.2025
# Last Modified By:    Jannis Necker

from bisect import bisect_right
from contextlib import suppress
from typing import Any, Sequence

from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.content.DataPoint import DataPoint
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.types import ChannelId, DataPointId, StockId
from ampel.util.mappings import unflatten_dict
from astropy.table import Table
from pydantic import TypeAdapter
from timewise.io.stable_tap import StableTAPService
from timewise.query import QueryType
from timewise.tables.allwise_p3as_mep import allwise_p3as_mep
from timewise.types import TYPE_MAP


class ConcurrentUpdateError(Exception):
    """
    Raised when the t0 collection was updated during ingestion
    """

    ...


class TiMongoMuxer(AbsT0Muxer):
    """
    This class compares info between alert and DB so that only the needed info is ingested.
    It checks for duplicate datapoints.
    """

    # Standard projection used when checking DB for existing PPS/ULS
    projection: dict[str, int] = {
        "_id": 0,
        "id": 1,
        "tag": 1,
        "channel": 1,
        "stock": 1,
        "body.mjd": 1,
        "body.w1_flux": 1,
        "body.w1_flux_error": 1,
        "body.w1_mag": 1,
        "body.w1_mag_error": 1,
        "body.w2_flux": 1,
        "body.w2_flux_error": 1,
        "body.w2_mag": 1,
        "body.w2_mag_error": 1,
        "body.ra": 1,
        "body.dec": 1,
    }

    channel: None | ChannelId | AnyOf[ChannelId] | AllOf[ChannelId] = None

    unique_key: list[str] = ["mjd", "ra", "dec"]

    # URL of tap service for query of AllWISE Source Table
    tap_service_url: str = "https://irsa.ipac.caltech.edu/TAP"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # used to check potentially already inserted pps
        self._photo_col = self.context.db.get_collection("t0")
        self._projection_spec = unflatten_dict(self.projection)

        self._tap_service = StableTAPService(self.tap_service_url)

        self._allwise_source_cntr: list[str] = []
        self._not_allwise_source_cntr: list[str] = []

    def process(
        self, dps: list[DataPoint], stock_id: None | StockId = None
    ) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
        """
        :param dps: datapoints from alert
        :param stock_id: stock id from alert
        Attempt to determine which pps/uls should be inserted into the t0 collection,
        and which one should be marked as superseded.
        """
        # IPAC occasionally issues multiple subtraction candidates for the same
        # exposure and source, and these may be received in parallel by two
        # AlertConsumers.
        for _ in range(10):
            with suppress(ConcurrentUpdateError):
                return self._process(dps, stock_id)
        raise ConcurrentUpdateError(
            f"More than 10 iterations ingesting alert {dps[0]['id']}"
        )

    # NB: this 1-liner is a separate method to provide a patch point for race condition testing
    def _get_dps(self, stock_id: None | StockId) -> list[DataPoint]:
        if self.channel is not None:
            if isinstance(self.channel, ChannelId):
                channel_query: (
                    ChannelId | dict[str, Sequence[ChannelId | AllOf[ChannelId]]]
                ) = self.channel
            elif isinstance(self.channel, AnyOf):
                channel_query = {"$in": self.channel.any_of}
            elif isinstance(self.channel, AllOf):
                channel_query = {"$all": self.channel.all_of}
            else:
                # should not happen
                raise TypeError()
            _channel = {"channel": channel_query}
        else:
            _channel = {}
        query = {"stock": stock_id, **_channel}
        return list(self._photo_col.find(query, self.projection))

    def _check_cntrs(self, dps: Sequence[DataPoint]) -> None:
        # assemble query
        query_config = {
            "type": "by_allwise_cntr_and_position",
            "radius_arcsec": 10,
            "columns": ["cntr"],
            "constraints": [],
            "table": {"name": "allwise_p3as_psd"},
        }
        query: QueryType = TypeAdapter(QueryType).validate_python(query_config)

        # load datapoints into astropy table
        upload = Table([dp["body"] for dp in dps])
        upload["allwise_cntr"] = upload[allwise_p3as_mep.allwise_cntr_column]
        upload[query.original_id_key] = [dp["id"] for dp in dps]
        for key, dtype in query.input_columns.items():
            upload[key] = upload[key].astype(TYPE_MAP[dtype])
        for key in upload.colnames:
            if key not in query.input_columns:
                upload.remove_column(key)

        # run query
        self.logger.info("Querying AllWISE Source Table for MEP CNTRs ...")
        res = self._tap_service.run_sync(
            query.adql, uploads={query.upload_name: upload}
        )

        # update internal state
        res_cntr = res.to_table()["cntr"].astype(str)
        self._allwise_source_cntr.extend(list(res_cntr))
        self._not_allwise_source_cntr.extend(
            list(set(upload["allwise_cntr"].astype(str)) - set(res_cntr))
        )

    def _check_mep_allwise_sources(self, dps: Sequence[DataPoint]) -> list[DataPointId]:
        dps_with_unchecked_cntr = [
            dp
            for dp in dps
            if str(dp["body"][allwise_p3as_mep.allwise_cntr_column])
            not in self._allwise_source_cntr + self._not_allwise_source_cntr
        ]
        if len(dps_with_unchecked_cntr) > 0:
            self._check_cntrs(dps_with_unchecked_cntr)

        # compile list of invalid datapoint ids
        invalid_dp_ids = []
        for dp in dps:
            cntr = str(dp["body"][allwise_p3as_mep.allwise_cntr_column])
            if cntr in self._not_allwise_source_cntr:
                invalid_dp_ids.append(dp["id"])

        return invalid_dp_ids

    def _process(
        self, dps: list[DataPoint], stock_id: None | StockId = None
    ) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
        """
        :param dps: datapoints from alert
        :param stock_id: stock id from alert
        Attempt to determine which pps/uls should be inserted into the t0 collection,
        and which one should be marked as superseded.
        """

        # Part 1: gather info from DB and alert
        #######################################

        # New pps/uls lists for db loaded datapoints
        dps_db = self._get_dps(stock_id)

        # python set of ids of datapoints from DB
        ids_dps_db = {el["id"] for el in dps_db}

        # Create set with new datapoint ids from alert
        new_dps = [dp for dp in dps if dp["id"] not in ids_dps_db]
        ids_dps_alert = {el["id"] for el in new_dps}

        if len(ids_dps_alert) == 0:
            self.logger.debug(f"{str(stock_id)}: no new data points")
            return None, None

        # uniquify photopoints by mjd, ra, and dec.
        # make sure there are no duplicates
        unique_dps_ids: dict[tuple[float, float, float], list[DataPointId]] = {}

        for dp in dps_db + new_dps:
            # jd alone is not enough for matching pps because each time is associated with
            # two filters! Also, if there can be multiple sources within the same frame which
            # leads to duplicate MJD and FID. Check position in addition.
            key = tuple(dp["body"][k] for k in self.unique_key)

            if target := unique_dps_ids.get(key):
                # insert id in order
                idx = bisect_right(target, dp["id"])
                if idx == 0 or target[idx - 1] != dp["id"]:
                    target.insert(idx, dp["id"])
            else:
                unique_dps_ids[key] = [dp["id"]]

        # Part 2: Check that there are no duplicates and handle redundant AllWISE MEP data
        ##################################################################################

        invalid_dp_ids = []
        for key, simultaneous_dps in unique_dps_ids.items():
            dps_db_wrong = [dp for dp in dps_db if dp["id"] in simultaneous_dps]
            dps_wrong = [dp for dp in dps if dp["id"] in simultaneous_dps]
            msg = (
                f"stockID {str(stock_id)}: Duplicate photopoints at {key}!\nDPS from DB:"
                f"\n{dps_db_wrong}\nNew DPS:\n{dps_wrong}"
            )

            all_wrong_dps = dps_db_wrong + dps_wrong
            if len(simultaneous_dps) > 1:
                # if these datapoints come from the AllWISE MEP database, downloaded by timewise
                # there can be duplicates. Only the AllWISE CNTR can tell us which datapoints
                # should be used: the CNTR that appears in the AllWISE source catalog.
                if all(
                    [
                        ("TIMEWISE" in dp["tag"]) and ("allwise_p3as_mep" in dp["tag"])
                        for dp in all_wrong_dps
                    ]
                ):
                    self.logger.info(
                        f"{len(all_wrong_dps)} duplicate MEP datapoints found. Checking ..."
                    )
                    i_invalid_dp_ids = self._check_mep_allwise_sources(
                        dps_db_wrong + dps_wrong
                    )
                    self.logger.info(
                        f"Found {len(i_invalid_dp_ids)} invalid MEP datapoints."
                    )
                    invalid_dp_ids.extend(i_invalid_dp_ids)

                else:
                    raise RuntimeError(msg)

        # Part 3: Compile final lists of datapoints to insert and combine
        #################################################################

        # Difference between candids from the alert and candids present in DB
        ids_dps_to_insert = ids_dps_alert - ids_dps_db - set(invalid_dp_ids)
        dps_to_insert = [dp for dp in dps if dp["id"] in ids_dps_to_insert]
        dps_to_combine = [
            dp
            for dp in dps + dps_db
            if dp["id"] in ((ids_dps_alert | ids_dps_db) - set(invalid_dp_ids))
        ]
        self.logger.debug(
            f"Got {len(ids_dps_alert)} datapoints from alerts, "
            f"found {len(dps_db)} in DB, "
            f"{len(invalid_dp_ids)} invalid datapoints, "
            f"inserting {len(dps_to_insert)} datapoints, "
            f"combining {len(dps_to_combine)} datapoints"
        )

        return dps_to_insert, dps_to_combine

    def _project(self, doc, projection) -> DataPoint:
        out: dict[str, Any] = {}
        for key, spec in projection.items():
            if key not in doc:
                continue

            if isinstance(spec, dict):
                item = doc[key]
                if isinstance(item, list):
                    out[key] = [self._project(v, spec) for v in item]
                elif isinstance(item, dict):
                    out[key] = self._project(item, spec)
            else:
                out[key] = doc[key]

        return out  # type: ignore[return-value]



from typing import Iterable, List
import pandas as pd
import numpy as np
from ampel.types import StockId
from ampel.content.DataPoint import DataPoint


def datapoints_to_dataframe(
    datapoints: Iterable[DataPoint],
    columns: list[str],
    check_tables: list[str] | None = None,
) -> tuple[pd.DataFrame, List[List[StockId]]]:
    """
    Convert a list of Ampel DataPoints into a pandas DataFrame.

    Parameters
    ----------
    datapoints : list of dict
        List of DataPoints (each must have a "body" dict).
    columns : list of str
        Keys from datapoint["body"] to include as DataFrame columns.
    check_tables: list of str
        check if the tables are in the tags

    Returns
    -------
    pd.DataFrame
        DataFrame with one row per DataPoint and one column per requested key.
    """

    records = []
    ids = []
    stock_ids = []
    for dp in datapoints:
        body = dp.get("body", {})
        # Build one row with only requested keys
        row = {col: body[col] for col in columns}
        # check if the tables are in tags
        if check_tables is not None:
            for table in check_tables:
                row[table] = table in dp["tag"]
        # build the index from datapoint ids
        ids.append(dp["id"])
        records.append(row)
        stock_ids.append(np.atleast_1d(dp["stock"]).tolist())

    colnames = columns if check_tables is None else columns + check_tables
    return pd.DataFrame.from_records(records, columns=colnames, index=ids), stock_ids



import logging
from typing import Dict, Literal

from .base import Query

logger = logging.getLogger(__name__)


class AllWISECntrQuery(Query):
    type: Literal["by_allwise_cntr_and_position"] = "by_allwise_cntr_and_position"
    radius_arcsec: float

    @property
    def input_columns(self) -> Dict[str, str]:
        return {
            "allwise_cntr": "int",
            "ra": "float",
            "dec": "float",
            self.original_id_key: "int",
        }

    def build(self) -> str:
        logger.debug(f"constructing query by AllWISE cntr for {self.table.name}")

        q = "SELECT \n\t"
        for k in self.columns:
            q += f"{self.table.name}.{k}, "
        q += f"\n\tmine.{self.original_id_key} \n"
        q += f"FROM\n\tTAP_UPLOAD.{self.upload_name} AS mine \n"
        q += f"RIGHT JOIN\n\t{self.table.name} \n"
        q += "WHERE \n"
        q += (
            f"\tCONTAINS(POINT('J2000',{self.table.name}.{self.table.ra_column},{self.table.name}.{self.table.dec_column}),"
            f"CIRCLE('J2000',mine.ra,mine.dec,{self.radius_arcsec / 3600:.18f}))=1 "
        )

        constraints = self.constraints + [
            f"{self.table.allwise_cntr_column} = {self.upload_name}.allwise_cntr"
        ]

        if len(constraints) > 0:
            q += " AND (\n"
            for c in constraints:
                q += f"\t{self.table.name}.{c} AND \n"
            q = q.strip(" AND \n")
            q += "\t)"

        logger.debug(f"\n{q}")
        return q



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                timewise/ampel/timewise/alert/TimewiseAlertSupplier.py
# License:             BSD-3-Clause
# Author:              Jannis Necker <jannis.necker@gmail.com>
# Date:                16.09.2025
# Last Modified Date:  16.09.2025
# Last Modified By:    Jannis Necker <jannis.necker@gmail.com>

import sys
from hashlib import blake2b
from typing import Literal, List, Dict, Any

import pandas as pd

from bson import encode

from ampel.alert.AmpelAlert import AmpelAlert
from ampel.base.AmpelABC import AmpelABC
from ampel.alert.BaseAlertSupplier import BaseAlertSupplier
from ampel.view.ReadOnlyDict import ReadOnlyDict


class TimewiseAlertSupplier(BaseAlertSupplier, AmpelABC):
    """
    Iterable class that, for each transient name provided by the underlying alert_loader
    returns a PhotoAlert instance.
    """

    stat_pps: int = 0
    stat_uls: int = 0

    dpid: Literal["hash", "inc"] = "hash"
    #    external_directory: Optional[ str ]
    #    deserialize: None | Literal["avro", "json"]

    bands: List[str] = ["w1", "w2"]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.counter = 0 if self.dpid == "hash" else 1

    def __next__(self) -> AmpelAlert:
        """
        :returns: a dict with a structure that AlertProcessor understands
        :raises StopIteration: when alert_loader dries out.
        :raises AttributeError: if alert_loader was not set properly before this method is called
        """
        table: pd.DataFrame = self._deserialize(next(self.alert_loader))  # type: ignore

        stock_ids = table["stock_id"].unique()
        assert len(stock_ids) == 1
        stock_id = stock_ids[0]

        # make the tables into a list of dictionaries for ampel to understand
        all_ids = b""
        pps = []

        # remove the _ep at the end of AllWISE MEP data
        columns_to_rename = [c for c in table.columns if c.endswith("_ep")]
        if len(columns_to_rename):
            rename = {
                c: c.replace("_ep", "")
                for c in columns_to_rename
                if c.replace("_ep", "") not in table.columns
            }
            if rename:
                # in this case only the allwise column eith the _ep extension exists
                # and we can simply rename the columns
                table.rename(columns=rename, inplace=True)

            move = {
                c: c.replace("_ep", "")
                for c in table.columns
                if (c.replace("_ep", "") in table.columns) and (c.endswith("_ep"))
            }
            if move:
                # In this case, the columns already exists because the neowise data is present
                # we have to insert the values form the columns with the _ep extension into the
                # respective neowise columns
                for c, nc in move.items():
                    na_mask = table[nc].isna()
                    table.loc[na_mask, nc] = table[c][na_mask]
                pd.options.mode.chained_assignment = None
                table.drop(columns=[c for c in move], inplace=True)
                pd.options.mode.chained_assignment = "warn"

        for i, row in table.iterrows():
            # convert table row to dict, convert data types from numpy to native python
            # Respect masked fields and convert to None
            pp = {str(k): None if pd.isna(v) else v for k, v in row.to_dict().items()}
            pp_hash = blake2b(encode(pp), digest_size=7).digest()
            if self.counter:
                pp["candid"] = self.counter
                self.counter += 1
            else:
                pp["candid"] = int.from_bytes(pp_hash, byteorder=sys.byteorder)

            all_ids += pp_hash
            pps.append(ReadOnlyDict(pp))

        if not pps:
            return self.__next__()

        # Update stats
        self.stat_pps += len(pps)

        return AmpelAlert(
            id=int.from_bytes(  # alert id
                blake2b(all_ids, digest_size=7).digest(), byteorder=sys.byteorder
            ),
            stock=int(stock_id),  # internal ampel id
            datapoints=tuple(pps),
        )



import logging
from typing import Annotated, Literal, List
from pathlib import Path

import typer

from .config import TimewiseConfig
from .plot.diagnostic import make_plot
from .process.interface import AMPEL_EXISTS

from rich.logging import RichHandler


ampel_missing_msg = (
    " (download only because AMPEL is not installed)" if not AMPEL_EXISTS else ""
)
app = typer.Typer(help="Timewsie CLI" + ampel_missing_msg)

config_path_type = Annotated[
    Path, typer.Argument(help="Pipeline config file (YAML/JSON)")
]
ampel_config_path_type = Annotated[Path, typer.Argument(help="AMPEL config YAML")]


# --- Global callback (runs before every command) ---
@app.callback()
def main(
    ctx: typer.Context,
    log_level: str = typer.Option(
        "INFO",
        "--log-level",
        "-l",
        help="Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
        case_sensitive=False,
    ),
):
    """Global options for all Timewise commands."""
    # Normalize log level
    level = getattr(logging, log_level.upper(), None)
    if not isinstance(level, int):
        raise typer.BadParameter(f"Invalid log level: {log_level}")

    # Rich logging
    logging.basicConfig(
        handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    )
    logging.getLogger("timewise").setLevel(level)

    # Store log level in context for subcommands
    ctx.obj = {"log_level": level}


@app.command(help="Download WISE photometry from IRSA")
def download(
    config_path: config_path_type,
    resubmit_failed: Annotated[
        bool,
        typer.Option(
            help="Re-submit jobs when failed due to connection issues",
        ),
    ] = False,
):
    TimewiseConfig.from_yaml(config_path).download.build_downloader(
        resubmit_failed=resubmit_failed
    ).run()


# the following commands will only be added if ampel is installed
if AMPEL_EXISTS:

    @app.command(help="Prepares the AMPEL job file so AMPEL can be run manually")
    def prepare_ampel(
        config_path: config_path_type,
    ):
        cfg = TimewiseConfig.from_yaml(config_path)
        ampel_interface = cfg.build_ampel_interface()
        p = ampel_interface.prepare(config_path)
        typer.echo(f"AMPEL job file: {p}")

    @app.command(help="Processes the lightcurves using AMPEL")
    def process(
        config_path: config_path_type,
        ampel_config_path: ampel_config_path_type,
    ):
        cfg = TimewiseConfig.from_yaml(config_path)
        ampel_interface = cfg.build_ampel_interface()
        ampel_interface.run(config_path, ampel_config_path)

    @app.command(help="Write stacked lightcurves to disk")
    def export(
        config_path: config_path_type,
        output_directory: Annotated[Path, typer.Argument(help="output directory")],
        indices: Annotated[
            list[int] | None,
            typer.Option(
                "-i", "--indices", help="Indices to export, defaults to all indices"
            ),
        ] = None,
    ):
        TimewiseConfig.from_yaml(config_path).build_ampel_interface().export_many(
            output_directory, indices
        )

    @app.command(help="Run download, process and export")
    def run_chain(
        config_path: config_path_type,
        ampel_config_path: ampel_config_path_type,
        output_directory: Annotated[Path, typer.Argument(help="output directory")],
        indices: Annotated[
            list[int] | None,
            typer.Option(
                "-i", "--indices", help="Indices to export, defaults to all indices"
            ),
        ] = None,
    ):
        download(config_path)
        process(config_path, ampel_config_path)
        export(config_path, output_directory, indices)

    @app.command(help="Make diagnostic plots")
    def plot(
        config_path: config_path_type,
        indices: Annotated[
            List[int],
            typer.Argument(help="Identifiers of the objects for which to create plots"),
        ],
        output_directory: Annotated[Path, typer.Argument(help="Output directory")],
        cutout: Annotated[
            Literal["sdss", "panstarrs"],
            typer.Option("-c", "--cutout", help="Which survey to use for cutouts"),
        ] = "panstarrs",
    ):
        make_plot(
            config_path,
            indices=indices,
            cutout=cutout,
            output_directory=output_directory,
        )



#!/usr/bin/env python
# File:                Ampel-ZTF/ampel/ztf/ingest/ZiCompilerOptions.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.05.2021
# Last Modified Date:  14.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.base.AmpelABC import AmpelABC


class TiCompilerOptions(CompilerOptions, AmpelABC):
    stock: dict[str, Any] = {"tag": "TIMEWISE"}
    t0: dict[str, Any] = {"tag": "TIMEWISE"}
    t1: dict[str, Any] = {"tag": "TIMEWISE"}
    state_t2: dict[str, Any] = {"tag": "TIMEWISE"}
    point_t2: dict[str, Any] = {"tag": "TIMEWISE"}
    stock_t2: dict[str, Any] = {"tag": "TIMEWISE"}



# keys for stacked lightcurve processing
MEAN = "mean"
MEDIAN = "median"
RMS = "rms"
UPPER_LIMIT = "ul"
NPOINTS = "npoints"
ZEROPOINT_EXT = "zeropoint"
FLUX_DENSITY_EXT = "fluxdensity"
KSTEST_NORM_EXT = "kstest_norm_pvalue"

# keys of the single exposure lightcurve
FLUX_EXT = "flux"
MAG_EXT = "mpro"
ERROR_EXT = "sig"



from typing import Annotated, TypeAlias, Union

from pydantic import Field

from .by_allwise_cntr_and_position import AllWISECntrQuery
from .positional import PositionalQuery

# Discriminated union of all query types
QueryType: TypeAlias = Annotated[
    Union[PositionalQuery, AllWISECntrQuery], Field(discriminator="type")
]
